#!/usr/bin/perl

require v5.8.4; # version 5.8.2 crashes in File::copy when running for too long

$version = "0.0.3";

sub usageInfo()
{
	print "\n";
	print "kbk.pl [-l|-ll <set>|-v|-vv <set>|-c ...|-b|-bb <set>]\n";
	print "\n";
	print "usage:\n";
	print " -l   list all backup sets\n";
	print " -ll <set>  list a specific backup set\n";
	print " -V   version num\n";
	print " -v   verify all backup sets, check for integrity\n";
	print " -vv <set>  verify a specific set, output a report.\n";
	print " -c  <set> <path> <dest> <date|now> copy path from the database to <dest>\n";
	print " -b  perform incremental backup on all sets\n";
	print " -bb <set> perform incremental backup on specific set\n";
	print "\n";
}

#################################
# parse command line.
my %cmdargs = ();
my $cmdline = "";
my $cmd_mode = "help", $cmd_set = "all", $arg0 = "", $arg1 = "", $arg2 = "", $arg3 = "";
sub readCommandLine()
{
	my $arg = 0;
	for ($arg = 0; $arg <= $argc; $arg++)
  	{
		$cmdline .= $ARGV[$arg] . " ";
	}

	if ($cmdline =~ m/-l/)
	{
		$cmd_mode = "list";
		if ($cmdline =~ m/-ll ([^-]\S*)/)
		{
			$cmd_set = $1;
		}
	}
	elsif ($cmdline =~ m/-v/)
	{
		$cmd_mode = "verify";
		if ($cmdline =~ m/-vv ([^-]\S*)/)
		{
			$cmd_set = $1;
		}
	}
   elsif ($cmdline =~ m/-V/)
	{
		$cmd_mode = "ver";
	}
	elsif ($cmdline =~ m/-c ([^-]\S*) ([^-]\S*) ([^-]\S*) ([^-]\S*)/)
	{
		$cmd_mode = "get";
		$cmd_set = $1;
		$arg0 = $1;
		$arg1 = $2;
		$arg2 = $3;
		$arg3 = $4;
	}
   elsif ($cmdline =~ m/-c ([^-]\S*) ([^-]\S*) ([^-]\S*)/) # same as prev, w/ opt arg4
	{
		$cmd_mode = "get";
		$cmd_set = $1;
		$arg0 = $1;
		$arg1 = $2;
		$arg2 = $3;
		$arg3 = "now";
	}
	elsif ($cmdline =~ m/-b/)
	{
		$cmd_mode = "backup";
		if ($cmdline =~ m/-bb ([^-]\S*)/)
		{
			$cmd_set = $1;
		}
	}
}
#######################################################################

# $parttostrip, $pathname 
sub stripPath( $$ )
{
	my $parttostrip = shift;
	my $path = shift;
	$path =~ s/$parttostrip//i;
	return $path;
}

# for mkpath(path, bool, perm)
use File::Path;
use File::Copy;

# create a directory tree on the filesystem
# args: tree
# return 0 if fail
sub makePath( $ )
{
   my $tree = shift;
   eval { mkpath( $tree, 0, 0777 ) };
   if ($@)
   {
      print STDOUT "\n[mkpath] Couldn't create  $tree: $@\n";
      return 0;
   }
   return 1;
}

# sub sig_gdb - send a CTRL_C to the debugged program {{{
sub sig_c
{
   print "***[ouch!  wait until done with current operation] ***.\n";
};

sub disableCtrlC()
{
   $SIG{INT}=\&sig_c; # disable CTRL-C
}
sub enableCtrlC()
{
   $SIG{INT}='DEFAULT';# restore CTRL-C
}

sub isWindowsPath( $ )
{
	my $path = shift;
	if ($path =~ m/^[A-Za-z]:[\\\/].*$/ || $path =~ m/^\\\\.*$/)  # c: or smb network..
	{
		return 1;
	}
	else
	{
		return 0;
	}
}

sub isSshPath( $ )
{
	my $path = shift;
	if ($path =~ m/^[^@]*@[^\.]*\.[^:]*:[\/].*$/)
	{
		return 1;
	}
	else
	{
		return 0;
	}
}

# indented printing...
my $iprint_indents = 0;
sub iprint( $ )
{
	my $stuff = shift;
	my $x = 0;
	for ($x = 0; $x < $iprint_indents; $x++)
	{
		print "  ";
	}
	print $stuff;
}
sub iprint_inc() { $iprint_indents++; }
sub iprint_dec() { $iprint_indents--; }

$spinner_index = 0;
sub spinner()
{
	@spins = ( '/', '|', '\\', '-', '/', '|', '\\', '-', '/', '|', '\\', '-', '/', '|', '\\', '-', 
					'_', '-', '*', '|', '^', '_', '-', '*', '|', '^', '_', '-', '*', '|', '^' );
	my $spinlength = @spins;
	$spinner_index = ($spinner_index + 1) % $spinlength;
	return $spins[$spinner_index];
}

sub justFilename( $ )
{
	# simple filename, no path info.
	my $filename = shift;
	$filename =~ s/.*[\/\\]([^\/\\]*)$/$1/g; # strip off the leading dir path
	return $filename;
}

sub justPath( $ )
{
	# simple filename, no path info.  
	my $filename = shift;
	$filename =~ s/(.*[\/\\])[^\/\\]*$/$1/g; # strip off the leading dir path
	return $filename;
}

sub getEnv( $ )
{
	$name = shift;
	$env = $ENV{$name};
	print "found $name == '$env'\n";
	if ($env eq "")
	{
		print "[error] set your $name env variable\n";
	}
	return $env;
}

sub reduceDotDots( $ )
{
	my $name = shift;
	while ($name =~ s/[\\\/][^\\\/]*[\\\/]\.\.//)
	{
	}
	return $name;
}
sub pad( $$ )
{
	my $num = shift;
	my $pads = shift;
	return sprintf("%0".$pads."d", $num);
}

sub getCurrentTimeString()
{
	my ($sec,$min,$hour,$mday,$mon,$year,
       $wday,$yday,$isdst) = localtime time;
	return (1900+$year).".".pad($mon+1,2).".".pad($mday,2)."-".pad($hour,2).".".pad($min,2).".".pad($sec,2);
}

sub baseName( $$ ) # yeah, hacky.
{
	my $basename = shift;
	my $length = shift;
	if ($length eq 4)
	{
		$basename =~ s/.*[\/\\]([^\/\\]*)[\/\\]([^\/\\]*)[\/\\]([^\/\\]*)[\/\\]([^\/\\]*)$/$1\/$2\/$3\/$4/g; # strip off the leading dir path
		return $basename;
	}
	if ($length eq 3)
	{
		$basename =~ s/.*[\/\\]([^\/\\]*)[\/\\]([^\/\\]*)[\/\\]([^\/\\]*)$/$1\/$2\/$3/g; # strip off the leading dir path
		return $basename;
	}
}


########################################################
## recursion module.
use File::Find;
my @findFunc_CallbackStack;
my $findFunc_CallbackStackPtr = -1;
sub findFunc() {
 $func = @findFunc_CallbackStack[$findFunc_CallbackStackPtr];
 if (-d $File::Find::name) {
	&$func( $File::Find::name, "dir" );
 } 
 elsif (-f $File::Find::name) {
	&$func( $File::Find::name, "file" );
 } 
}
## recurse interface
## call this like so:  recurseDirs( cwd, \&{"main::myRecurseFunc"} );
sub recurseDirs( $$ ) {
	@current_dir = shift;
	$findfunc_callback = shift;
	$findFunc_CallbackStackPtr += 1;
	@findFunc_CallbackStack[$findFunc_CallbackStackPtr] = $findfunc_callback;
	find( \&findFunc, @current_dir );
	$findFunc_CallbackStackPtr -= 1;
}
## end of recursion module.
########################################################


use Cwd;
$dir = cwd;
$root_dir = cwd;

$argc = $#ARGV;
if ($#ARGV == 1)
{
	#todo: process args here....
}

###############################
# return the md5sum for a given filename
use Digest::MD5;
sub getMD5( $ )
{
	my $file = shift;
	if (open( MD5FILE, $file ))
	{
      binmode( MD5FILE );
      my $ctx = Digest::MD5->new;
      $ctx->addfile( *MD5FILE );
		my $hexdigest = $ctx->hexdigest;
      close( MD5FILE );
		return $hexdigest;
	}
	else
	{
		print "[md5] Can't open '$file': $!\n";
	}
	return 0;
}
sub validatedCopy( $$ )
{
	my $source = shift;
	my $dest = shift;
	my $s = getMD5( $source );
	copy( $source, $dest );
	my $d = getMD5( $dest );
	if ($s ne $d)
	{
		return 0;
	}
	return 1;
}

sub wildcardExpToRegExp( $ )
{
	my $exp = shift;
	$exp =~ s/\\/\//g; # standardize on unix style slashes.
	$exp =~ s/\./\\./g;
	$exp =~ s/\*/.*/g;
	$exp =~ s/\?/./g;
	$exp = "^" . $exp . "\$";
	return $exp;
}

# debug function.
# pass in an array by reference. like this  \@array
sub outputArray( @ )
{
	my $exps = shift;
	foreach my $item (@{ $exps }) # cast scalar pointer to actual array.
	{
		print " - $item\n";
	} 
}

%global = ();
%sets = ();
my $errorlog = $root_dir . "/errorlog.txt";
my $statuslog = $root_dir . "/statuslog.txt";
sub readConfig( $ )
{
	my $filename = shift;
	open( INHANDLE, "< $filename" );
	my @text = <INHANDLE>;
  	close( INHANDLE );
	my $cur_setname = "";
	my $indent_level = 0;
	foreach $line (@text)
	{
		if ($line =~ m/}/)
		{
			$indent_level = $indent_level - 1; 
			$cur_setname = ""; # done with block.
			$in_global = 0;   # done...
		}
		elsif ($line =~ m/{/)
		{
			$indent_level = $indent_level + 1;
		}

		# detect a set tag
		if ($indent_level == 0 && $line =~ m/set\s+"([^"]+)"/i)
		{
			$cur_setname = $1;
			my %hash = ();
			$sets{$cur_setname} = \%hash;
		}

		# detect a set parameter
		if ($cur_setname ne "" && $indent_level > 0)
		{
			if ($line =~ m/([^\s=]+)\s*=\s*["]([^"]*)["];/)
			{
				my $key = $1, $value = $2;
				$sets{$cur_setname}->{ $key } = $value;
				if (isWindowsPath( $sets{ $cur_setname }->{ $key } ))
				{
					$sets{ $cur_setname }->{ $key } =~ s/\\/\//g;
				}
				$sets{ $cur_setname }->{ $key } =~ s/(.*)\/$/$1/g;
			}
		}

		# detect a global tag
		if ($indent_level == 0 && $line =~ m/global/i)
		{
			$in_global = 1;
		}

		# detect a global parameter
		if ($in_global >= 0 && $indent_level > 0)
		{
			if ($line =~ m/([^\s=]+)\s*=\s*["]([^"]*)["];/)
			{
				my $key = $1, $value = $2;
				$global{ $key } = $value;
				if (isWindowsPath( $global{ $key } ))
				{
					$global{ $key } =~ s/\\/\//g;
				}
			}
		}
	}

	# test:
	foreach my $item (keys %sets)
	{
		# apply defaults for fields that are not set.
		if ($sets{$item}->{"excludetype"} eq "")
		{
			$sets{$item}->{"excludetype"} = "wildcard";
		}
		if ($sets{$item}->{"includetype"} eq "")
		{
			$sets{$item}->{"includetype"} = "wildcard";
		}
		
		# split up the include and exclude file sets...
		if ($sets{$item}->{"include"} ne "")
		{
			my @includes = (); # new array.
			@includes = split( /[;]/, $sets{$item}->{"include"} );
			$sets{$item}->{"includes"} = \@includes; 
		}
		if ($sets{$item}->{"exclude"} ne "")
		{
			my @excludes = (); # new array.
			@excludes = split( /[;]/, $sets{$item}->{"exclude"} );
			$sets{$item}->{"excludes"} = \@excludes;
		}

		# convert simple wildcard expressions to regexp's we can use later...
		if ($sets{$item}->{"includetype"} eq "wildcard")
		{
			foreach my $item (@{ $sets{$item}->{"includes"} })
			{
				$item = wildcardExpToRegExp( $item );
			}
		}
		if ($sets{$item}->{"excludetype"} eq "wildcard")
		{
			foreach my $item (@{ $sets{$item}->{"excludes"} })
			{
				$item = wildcardExpToRegExp( $item );
			}
		}
		
		# output
		print "\n";
		print "[set \"$item\"]\n";
		print "description = \"" . $sets{$item}->{"description"} . "\"\n";
		print "source = \"" . $sets{$item}->{"source"} . "\"\n";
		print "dest = \"" . $sets{$item}->{"dest"} . "\"\n";
		#print "include = \"" . $sets{$item}->{"include"} . "\"\n";
		foreach my $item (@{ $sets{$item}->{"includes"} })
		{
			print "include = \"" . $item . "\"\n";	 
		}
		#print "exclude = \"" . $sets{$item}->{"exclude"} . "\"\n";
		foreach my $item (@{ $sets{$item}->{"excludes"} })
		{
			print "exclude = \"" . $item . "\"\n";	 
		}
	}

	print "\n";
	print "[global]\n";
	print "tempdir = \"" . $global{"tempdir"} . "\"\n";
	if (!-d $global{"tempdir"})
   {
      print "temp dir not found, creating " . $global{"tempdir"} . "\n";
      if (!makePath( $global{"tempdir"} ))
      {
         print "[error] couldn't create tempdir: " . $global{"tempdir"} . "\n";
         exit( 0 ); # critical
      }
   }
   $errorlog = $global{"errorlog"};
   $statuslog = $global{"statuslog"};
   print "statuslog = \"" . $statuslog . "\"\n";
   print "errorlog = \"" . $errorlog . "\"\n";
}

 sub autoflush (*$) {
# Set the state of autoflush for
# the handle supplied as $_[0]
# to the boolean state supplied as $_[1]
    select(
        do{
            my $old = select( $_[0] );
            $| = $_[1];
            $old;
        }
    )
}

sub logError( $ )
{
   if (open( ERRORFILE, ">> $errorlog" ))
   {
      my $text = shift;
      print ERRORFILE $text;
      autoflush(ERRORFILE, 1);
      close( ERRORFILE );
   }
}

sub logStatus( $ )
{
   my $text = shift;
   print STDOUT $text;
   autoflush(STDOUT, 1);
   if (open( STATUSFILE, ">> $statuslog" ))
   {
      print STATUSFILE $text;
      autoflush(STATUSFILE, 1);
      close( STATUSFILE );
   }
}

sub ilogError( $ )
{
   if (open( ERRORFILE, ">> $errorlog" ))
   {
      my $text = shift;
      print ERRORFILE $text;
      autoflush(ERRORFILE, 1);
      close( ERRORFILE );
   }
}

sub ilogStatus( $ )
{
   my $text = shift;
   iprint $text;
   autoflush(STDOUT, 1);
   if (open( STATUSFILE, ">> $statuslog" ))
   {
      print STATUSFILE $text;
      autoflush(STATUSFILE, 1);
      close( STATUSFILE );
   }
}

sub logErrorAndStatus( $ )
{
   my $text = shift;
   logStatus( $text );
   logError( $text );
}

sub ilogErrorAndStatus( $ )
{
   my $text = shift;
   ilogStatus( $text );
   ilogError( $text );
}

sub logFlush()
{
   STDOUT->autoflush(1);
}

sub my_handler
{
   my($sig) = @_;
   logErrorAndStatus "\n!!!Caught a SIG$sig--shutting down!!!\n";
}
#use sigtrap qw(handler my_handler any normal-signals handler my_handler any error-signals handler my_handler any old-interface-signals);

# self tests:
isSshPath( "kevin\@blah.blah.com:/my/path/to/joy" ) or die "failed ssh test\n";
0 eq isSshPath( "c:\\blah\\blah\\blah" ) or die "failed ssh test\n";
isWindowsPath( "c:\\blah\\blah\\blah" ) or die "failed windows test\n";
isWindowsPath( "\\\\192.168.1.121\\root-r\\" ) or die "failed samba test\n";
0 eq isWindowsPath( "kevin\@blah.blah.com:/my/path/to/joy" ) or die "failed windows test\n";
getDateValueForFile( "E:/backups/testing/cri/PSX2/mwlib/include/cri_adxt.h/2004.04.06-22.03.59.full" ) eq "20040406220359" or die "failed test getDateValueForFile\n";
getDateValueForFile( "2004.04.06-22.03.59.full" ) eq "20040406220359" or die "failed test getDateValueForFile\n";
justPath( "aki/bok/" ) eq "aki/bok/" or die "failed just path test\n";
justPath( "aki/bok/filename.gz" ) eq "aki/bok/" or die "failed just path test\n";
#open( TESTFILE, "> b5e9186a08f6d5685e4e21023d792b35.test" );
#print TESTFILE "some text to hash\n";
#close( TESTFILE );
#"b5e9186a08f6d5685e4e21023d792b35" eq getMD5( "b5e9186a08f6d5685e4e21023d792b35.test" ) or die "failed md5sum test\n";
#unlink( "b5e9186a08f6d5685e4e21023d792b35.test" );

####################################################3
# start.
$datestring = getCurrentTimeString();
readCommandLine();
readConfig( "config.ini" );

logError "======================================\n";
logErrorAndStatus "Command line: \"$cmdline\"\n";
logErrorAndStatus "Running in mode[$cmd_mode] with set[$cmd_set]\n";
logErrorAndStatus "Current time is: $datestring\n\n";
logError "======================================\n";



my $set;

# called for every file
# return 1 for success, 0 for failure.
sub copyTreeStructureFunc( $$ )
{
	my $key = shift;
	my $type = shift; # "dir" or "file"
	
	my $source = $sets{$set}->{"source"};
	my $dest = $sets{$set}->{"dest"};
	
	# strip off the source name
	$key =~ s/$source\/?//i;
   return makePath( $dest . "/" . $key );
}

my @file_listing = ();
sub getFileListingFunc( $$ )
{
	my $key = shift;
	my $type = shift; # "dir" or "file"
	if ($type eq "file")
	{
		@file_listing = ( @file_listing, $key );
	}
}
sub getFileListing( $ )
{
	my $searchpath = shift;
	@file_listing = ();
	recurseDirs( $searchpath, \&{"main::getFileListingFunc"} );
	return \@file_listing;
}
sub getDateValueForFile( $ )
{
	my $item = shift;
	if ($item =~ s/.*?[\/\\]?(.?.?.?.?)\.(.?.?)\.(.?.?)-(.?.?)\.(.?.?)\.(.?.?)\..*/$1$2$3$4$5$6/i)
	{
		return scalar $item;
	}
	elsif ($item =~ s/.*?[\/\\]?(.?.?.?.?)\.(.?.?)\.(.?.?)\..*/$1$2$3$4$5$6/i)
	{
		return scalar ($item . "000000"); #time...
	}
	return 0;
}

sub getDateValue( $ )
{
	my $item = shift;
	# year.mo.da-hr.mi.se
	if ($item =~ s/(.?.?.?.?)\.(.?.?)\.(.?.?)-(.?.?)\.(.?.?)\.(.?.?)/$1$2$3$4$5$6/i)
	{
		return scalar $item;
	}
	# no time, just year-mo-da
	elsif ($item =~ s/(.?.?.?.?)\.(.?.?)\.(.?.?)/$1$2$3/i)
	{
		return scalar ($item . "000000"); #time...
	}
	return 0;
}

# from the database, get the newest (in time) entry for the specified file.
# arg - database absolute path (e:/db/myfile.gz) 
# arg - type  (patch, full, md5sum)
sub getNewestFile( $$ )
{		
	my $newest = "";
	my $searchpath = shift;
	my $type = shift;

	getFileListing( $searchpath );

	foreach $item (@file_listing)
	{
		if ($item =~ m/\.$type$/i &&
			 getDateValueForFile( $item ) > getDateValueForFile( $newest ))
		{
			$newest = $item;
		}
	}
	return $newest;
}

sub createMd5File( $ )
{
	my $filename = shift;
	my $md5 = getMD5( $filename );
   if ($md5 ne "")
   {
      if (open( MD5FILE, "> $filename.md5sum" ))
      {
         print MD5FILE $md5 . "\n";
         close( MD5FILE );
         return 1;
      }
   }
   return 0;
}

sub compareFiles( $$ )
{
	my $file1 = shift;
	my $file2 = shift;
	if (!-e $file1 || !-e $file2)
	{
		return 0; # not found
	}

	if (getMD5( $file1 ) eq getMD5( $file2 ))
	{
		return 1;
	}
	return 0;
}

# create a patch.
# depends on external utility bspatch.exe
# arg - newer file
# arg - output file
# arg - diff patch file from new to old
sub patch( $$$ )
{
   disableCtrlC();
   
	my $new = shift;
	my $output = shift;
	my $patch = shift;
   if (!-f "$root_dir/bin/bspatch.exe")
   {
      ilogErrorAndStatus " - [critical] couldn't find $root_dir/bin/bspatch, fix!\n";
      exit( 0 );
   }
   else
   {
      system( "$root_dir/bin/bspatch \"$new\" \"$output\" \"$patch\"" );
   }

   enableCtrlC();
}

# verify that
# arg - newer file
# arg - older file
# arg - diff patch file from new to old
# returns 1 if patch against new yields the old file.
sub verifyPatch( $$$ )
{
	my $new = shift;
	my $old = shift;
	my $patch = shift;
	my $output = $global{"tempdir"} . "/" . justFilename( $old );
   patch( $new, $output, $patch );
   #system( "$root_dir/bin/bspatch $new $output $patch" );
	my $success = compareFiles( $old, $output );
	unlink $output;
	return $success;
}

# create a patch.
# depends on external utility bspatch.exe
# arg - from file
# arg - to file
# arg - output filename diff patch file from new to old
sub createPatch( $$$ )
{
   disableCtrlC();
   
   my $from = shift;
	my $to = shift;
	my $patchoutput = shift;
   if (!-f "$root_dir/bin/bsdiff.exe")
   {
      ilogErrorAndStatus " - [critical] couldn't find $root_dir/bin/bsdiff, fix!\n";
      exit( 0 );
   }
   else
   {
      system( "$root_dir/bin/bsdiff \"$from\" \"$to\" \"$patchoutput\"" );
   }
   enableCtrlC();
}

# pass in relative path...  aki/my/file.gz
# return 0 for nochange, 1 for changed.
sub hasFileChanged( $ )
{
	my $source = $sets{$set}->{"source"};
	my $dest = $sets{$set}->{"dest"};
	
	my $filename = shift;
	my $dblatestfile = getNewestFile( "$dest/$filename", "full" );
	if (compareFiles( $dblatestfile, "$source/$filename" ))
	{
		# equal
		return 0; # no change
	}
	return 1;  # changed;
}

sub copyFile( $$ )
{
   disableCtrlC();
   
	my $src = shift;
	my $dest = shift;

   my $retval = 0;
   logStatus( "," );
   eval { $retval = File::Copy::syscopy( "$src", "$dest" ); };
   logStatus( "_" );
   my $exception = $@;
   if ($exception)
   {
      logErrorAndStatus "Exception: " . $exception . "\n";
      logErrorAndStatus "Exception msg (if any): " . $exception->{msg} . "\n";
      enableCtrlC();
      return 0;
   }
   enableCtrlC();
   return $retval;
}

# advanced copy function.
# copies src file to the dest file, if the dir tree doesn't exist, 
# it creates it.
# arg - src path to file
# arg - dest path to file  (or directory with trailing '/')
# return 1 for success 0 for failure.
sub installFile( $$ )
{
   disableCtrlC();
   
	my $src = shift;
	my $dest = shift;
   #logStatus( "." );
	if (makePath( justPath( $dest ) ))
   {
      my $retval = 0;
      #logStatus( "," );
	   eval { $retval = File::Copy::syscopy( "$src", "$dest" ); };
      #logStatus( "_" );
	   my $exception = $@;
      if ($exception)
      {
         logErrorAndStatus "Exception: " . $exception . "\n";
         logErrorAndStatus "Exception msg (if any): " . $exception->{msg} . "\n";
         enableCtrlC();
         return 0;
      }
	   enableCtrlC();
      return $retval;
   }
   else
   {
      logErrorAndStatus( ". makePath failed. " );
      enableCtrlC();
      return 0;
   }
}

sub verifyIntegrity( $ );

sub isInSet( $$ )
{
	my $name = shift;
	my $exps = shift;
	foreach my $item (@{ $exps })
	{
   	if ($name =~ m/$item/ig)
		{
			return 1;
		}
	}
	return 0;
}



# called for every file
sub backupFilesFunc( $$ )
{
	my $key = shift;
	my $type = shift; # "dir" or "file"
	
	my $tempdir = $global{"tempdir"};
	my $source = $sets{$set}->{"source"};
	my $dest = $sets{$set}->{"dest"};
	my @includes = @{ $sets{$set}->{"includes"} };
	my @excludes = @{ $sets{$set}->{"excludes"} };

	if ($type eq "file" && isInSet( $key, \@includes ) && !isInSet( $key, \@excludes ))
	{
		# get the filepath common to the db and source trees
		my $common_path = $key;
		$common_path =~ s/$source\/?//i;  # aki/myfile.gz   (no c:/home/)

		# get the filename without the path
		my $filename = justFilename( $common_path );

		# if the db already has the latest file, there is nothing to do
		if (!hasFileChanged( $common_path ))
		{
			#ilogStatus "no change in $common_path\n";		
			return;
		}

		# the database doesn't have the file, back it up!
		ilogStatus "backing up $common_path: ";
	
		# find most recent 'full' file in the database 
		# (before we update it with a newer one!):
		my $last_file = getNewestFile( "$dest/$common_path", "full" );

		# copy in the latest 'full' file	
		my $db_filename = "$datestring.full";
		ilogStatus "[copying";
		if (!installFile( "$source/$common_path", "$dest/$common_path/$db_filename" ))
		{
			logStatus " failed (this is bad)]\n";
			logError "backing up $common_path: [copying failed (this is bad)]\n";
			return;
		}
      logStatus ",";
      
		# output a marker.
      if (!-f "$dest/$common_path/db")
      {
		   logStatus "tag";
		   if (open( MARKERFILE, "> $dest/$common_path/db" )) # marker
         {
            print MARKERFILE "db";
		      close( MARKERFILE );
         }
      }

		# write the md5sum file.
		logStatus ",md5sum";
		if (!createMd5File( "$dest/$common_path/$db_filename" ))
      {
         logStatus ".. fail.. ";
      }
      logStatus "]\n";
		
		# if there already was a file in the db, we'll want to store an
		# incremental change (we store the minimal amount of information).
		# make a patch from latest file to the older one.
		if ($last_file ne "")
		{
			# filename for the diffpatch.
			my $patchfile = "$dest/$common_path/" . justFilename( $last_file );
			$patchfile =~ s/\.full$/.patch/i;
			
			# output a diffpatch from $db_filename to $last_file
			ilogStatus " - replacing old full version with incremental diff [patch";
         createPatch( "$dest/$common_path/$db_filename", "$last_file", "$patchfile" );
			if (!-f $patchfile)
			{
				ilogErrorAndStatus " - [error] couldn't create patch file $patchfile\n";
            ilogErrorAndStatus " - [info] command was 'createPatch(
               \"$dest/$common_path/$db_filename\", \"$last_file\",
               \"$patchfile\" );'\n"; 
			}
         # patch file worked, check it out...
         else
         {
   			#output patch's md5sum file
   			logStatus ",md5";
   			createMd5File( "$patchfile" );

   			# verify the patch worked...
   			logStatus ",verifying";
   			if (verifyPatch( "$dest/$common_path/$db_filename", "$last_file", "$patchfile" ))
   			{
   				logStatus ",rm old]\n";
   				unlink( $last_file );
   				# leave the full version's md5sum, we can use it later.
   			}
   			else
   			{
               # bad patch, remove it,
               # and undo the new full file as well,
               # as if we never touched the directory.
               unlink( $patchfile );
               unlink( "$patchfile.md5sum" );
               unlink( "$dest/$common_path/$db_filename" );
   				unlink( "$dest/$common_path/$db_filename.md5sum" );
   				ilogErrorAndStatus " failed]\n";
   				ilogErrorAndStatus " - [CRITICAL] couldn't make patch for $last_file (check db for corruption, i _think_ i backed out the operation ok, but check it please... :) )\n";
               exit( -1 );
            }
         }
		}
	}	
}

sub backup( $ )
{
	$set = shift;
	my $description = $sets{$set}->{"description"};
	my $source = $sets{$set}->{"source"};
	my $include = $sets{$set}->{"include"};
	my $exclude = $sets{$set}->{"exclude"};
	my $dest = $sets{$set}->{"dest"};

	if (-e $source)
	{
		recurseDirs( $source, \&{"main::backupFilesFunc"} );
	}
	else
	{	
		ilogStatus " - Path $source not found for set[$set].\n";
	}
}


# called for every file
sub listFilesFunc( $$ )
{
	my $key = shift;
	my $type = shift; # "dir" or "file"
	
	my $tempdir = $global{"tempdir"};
	my $source = $sets{$set}->{"source"};
	my $dest = $sets{$set}->{"dest"};
	
	if ($type eq "dir" && -f "$key/db")
	{
		# get the filepath common to the db and source trees
		my $common_path = $key;
		$common_path =~ s/$dest\/?//i;  # aki/myfile.gz   (no c:/home/)

		# get the filename without the path
		my $filename = justFilename( $common_path );
		
		if ($common_path eq "")
		{
			ilogStatus " $set:/.\n";
		}
		else
		{
			ilogStatus " $set:/$common_path\n";
		}
	}
}

sub list( $ )
{
	$set = shift;
	my $description = $sets{$set}->{"description"};
	my $source = $sets{$set}->{"source"};
	my $include = $sets{$set}->{"include"};
	my $exclude = $sets{$set}->{"exclude"};
	my $dest = $sets{$set}->{"dest"};

	if (-e $dest)
	{
		recurseDirs( $dest, \&{"main::listFilesFunc"} );
	}
	else
	{	
		iprint " - Path $dest not found for set[$set].\n";
	}
}

sub printDebugDbFile( $ )
{
	my $key = shift;
	my $tempdir = $global{"tempdir"};
	my $source = $sets{$set}->{"source"};
	my $dest = $sets{$set}->{"dest"};

	# get the filepath common to the db and source trees
	my $common_path = $key;
	$common_path =~ s/$dest\/?//i;  # aki/myfile.gz   (no c:/home/)

	# get the filename without the path
	my $filename = justFilename( $common_path );

	$file_arrayref = getFileListing( $key );
	ilogStatus "$common_path\n";
	foreach $item (@{$file_arrayref})
	{
		ilogStatus " - " . justFilename( $item ) . "\n";
	}
}

sub verifyChecksum( $$ )
{
	my $filename = shift;
	my $checksum = shift;
	
	# compare previously recorded md5sum to current md5sum
	if (open( MD5FILE, "< " . $checksum ))
	{
		my @bok = <MD5FILE>;
		my $previous_md5 = $bok[0];
		$previous_md5 =~ s/\n//g;
		close( MD5FILE );
		my $current_md5 = getMD5( $filename );
		if ($current_md5 ne 0 && $previous_md5 =~ m/$current_md5/)
		{
			#ilogStatus " - " . justFilename( $full_and_patches[$x] ) . " ok!\n";
		}
		else
		{
			ilogErrorAndStatus " - [fail!] \"$previous_md5\" != \"$current_md5\"... md5 doesn't match for " . justFilename( $checksum ) . " \n";
			return 0;
		}
	}
	else
	{
		ilogErrorAndStatus " - [error] couldn't open " . $checksum . "\n";
		return 0;
	}
	return 1;
}

# verify integrity of one file's entry in the database.
# also checks all past versions of the file.
sub verifyIntegrity( $ )
{
	my $key = shift;

	# check that all the db entries haven't been corrupted...
	$file_arrayref = getFileListing( $key );
	@full_and_patches = ();
	foreach $item (@{$file_arrayref})
	{
		if ($item =~ m/.full$/i || $item =~ m/.patch$/i)
		{
			@full_and_patches = ( @full_and_patches, $item );
		}
	}
	@full_and_patches = sort {$b cmp $a} @full_and_patches;
	my $x = 0;
	for ($x = 0; $x < @full_and_patches; $x++)
  	{
		# compare previously recorded md5sum to current md5sum
		if (!verifyChecksum( $full_and_patches[$x], $full_and_patches[$x] . ".md5sum" ))
		{
			return 0;
		}
	}
	return 1;
}



# called for every file
my $integrity_check = 0;
sub integrityFilesFunc( $$ )
{
	my $key = shift;
	my $type = shift; # "dir" or "file"

	if ($type eq "dir" && -f "$key/db")
	{
		#printDebugDbFile( $key );
		if (verifyIntegrity( $key ) eq 0)
		{
			$integrity_check = 1;
		}
	}
}

sub integrity( $ )
{
	$set = shift;
	my $description = $sets{$set}->{"description"};
	my $source = $sets{$set}->{"source"};
	my $include = $sets{$set}->{"include"};
	my $exclude = $sets{$set}->{"exclude"};
	my $dest = $sets{$set}->{"dest"};

	if (-e $dest)
	{
		recurseDirs( $dest, \&{"main::integrityFilesFunc"} );

		if ($integrity_check)
		{
			ilogErrorAndStatus " - [error!]  integrity check failed for set $set.\n";
		}
		else
		{
			ilogStatus " - [success] integrity check passed.\n";
		}
	}
	else
	{	
		ilogErrorAndStatus " - [error] Path $dest not found for set[$set].\n";
	}
}

sub isDbFile( $ )
{
	my $dblocation = shift;
	if (-f "$dblocation/db")
	{
		return 1;
	}
	return 0;
}

sub isDbDir( $ )
{
	return !isDbFile( shift );
}

sub interpretDate( $ )
{
	my $thedate = shift;
	$thedate =~ s/now/$datestring/i;
	return $thedate;
}

# convert ./blah/blah to $dir/blah/blah
sub getAbsolutePath( $$ )
{
   my $dir = shift;
   my $path = shift;
   
   # if absolute, do nothing, otherwise, append cwd
   if ($path =~ m/^[a-zA-Z]:[\\\/]/i ||
       $path =~ m/^[\\\/]/i)
   {}
   else
   {
      $path =~ s/^\.[\\\/]?//ig; # remove leading ./
      $path = join( "/", $dir, $path );
      $path =~ s/[\\\/]+$//ig; # remove trailing /
   }
   return $path;
}


$retrieveDirFunc_dblocation = "";
$retrieveDirFunc_destlocation = "";
$retrieveDirFunc_requested_date = "";
sub retrieveDirFunc( $$ )
{
	my $key = shift;
	my $type = shift; # "dir" or "file"
	if ($type eq "dir")
	{
      my $newdest = $key;
      $retrieveDirFunc_dblocation_minus_last = justPath( $retrieveDirFunc_dblocation );
      $newdest =~ s/$retrieveDirFunc_dblocation_minus_last//ig;
      $newdest = $retrieveDirFunc_destlocation . "/" . $newdest;
   if (isDbFile( $key ))
	{
      #$newdest = justPath( $newdest );
      #print "retrieving " . $key . "\n\t\t to " . $newdest . "\n";
      retrieve( $key,
                  $newdest,
                  $retrieveDirFunc_requested_date );
	}
   else
	{
      #print "retrieving DIR " . $key . "\n\t\t to " . $newdest . "\n";
      print "- path '$newdest'\n";
      makePath( $newdest );
	}
   }
}
sub retrieveDir( $$$ )
{
	$retrieveDirFunc_dblocation = shift;
	$retrieveDirFunc_destlocation = shift;
	$retrieveDirFunc_requested_date = shift;

   # remove trailing slash on dblocation
   $retrieveDirFunc_dblocation =~ s/[\\\/]+$//ig;

   # remove trailing slash on dest
   $retrieveDirFunc_destlocation =~ s/[\\\/]+$//ig;
   $retrieveDirFunc_destlocation = getAbsolutePath( $dir, $retrieveDirFunc_destlocation );

   if (isDbFile( $retrieveDirFunc_dblocation ) ||
       -f $retrieveDirFunc_dblocation)
   {
      print "error: entered retrieveDir on a file ($retrieveDirFunc_dblocation)... strange.\n";
      exit( 0 );
   }
	recurseDirs( $retrieveDirFunc_dblocation, \&{"main::retrieveDirFunc"} );
	return \@file_listing;
}

sub retrieve( $$$ )
{
	$set = $cmd_set;
	my $dblocation = shift;
	
	my $tempdir = $global{"tempdir"};
	my $source = $sets{$set}->{"source"};
	my $destlocation = shift;
	my $requested_date = shift;
	
	# check inputs...
	# see if file exists in the db...
	if (!-d $dblocation)
	{
		ilogErrorAndStatus " - [error] $dblocation not found in $cmd_set.\n";
		ilogErrorAndStatus " -         see $dblocation, check your paths.\n";
   }

	# gather all the db entries for the file.
	# ordered from recent to oldest
	$file_arrayref = getFileListing( $dblocation );
	my @full_and_patches = ();
	foreach $item (@{$file_arrayref})
	{
		if ($item =~ m/.full$/i || $item =~ m/.patch$/i)
		{
			@full_and_patches = ( @full_and_patches, $item );
		}
	}
	@full_and_patches = sort {$b cmp $a} @full_and_patches;

	# figure out how far back we need to go:
	# build a list of relevant files...
	my @getlist = ();
	my $x = 0;
	for ($x = 0; $x < @full_and_patches; $x++)
	{
		# if the date we're requesting is newer than current file
		# 4, 3, r, 2, 1 -> will yield 4 and 3
		if (getDateValueForFile( $full_and_patches[$x] ) ge getDateValue( $requested_date ))
		{
			@getlist = ( @getlist, $full_and_patches[$x] );
		}
		# 4, 3, r, 2, 1 -> will yield 2 only
		elsif (getDateValue( $requested_date ) gt getDateValueForFile( $full_and_patches[$x] ))
		{
			@getlist = ( @getlist, $full_and_patches[$x] );
			last; # break from the if, and the for
		}
		if (getDateValueForFile( $full_and_patches[$x] ) eq getDateValue( $requested_date ))
		{
			last;
		}
	}
	# special case
	# 4, 3, 2, 1, r
	if (@getlist gt 0 && $x eq @getlist && getDateValueForFile( $getlist[$x - 1] ) gt getDateValue( $requested_date ))
	{
		ilogErrorAndStatus " - [warning] file " . justFilename( $destlocation ) . " didn't exist at this time\n";
		return;
	}
	
	# if there is only 1 entry in the db, it is the full version, 
	# simply get it.
	if (@getlist eq 1)
	{
		if (verifyChecksum( $getlist[0], $getlist[0] . ".md5sum" ))
		{
			ilogStatus " - getting " . justFilename( $destlocation ) . "\n";
			copy( $getlist[0], $destlocation );
		}
		else
		{
			ilogErrorAndStatus " - can't get file " . $getlist[0] . ", checksum was invalid, possible corruption\n";
			return;
		}
	}

	# if there is more than one entry in the db, apply patches until we get 
	# the right date
	elsif (@getlist gt 1)
	{
		if (!verifyChecksum( $getlist[0], $getlist[0] . ".md5sum" ) || 
			 !verifyChecksum( $getlist[1], $getlist[1] . ".md5sum" ))
		{
			ilogErrorAndStatus " - can't use patch file " . $getlist[0] . ", checksum was invalid, possible corruption\n";
			return;
		}
		
		# make patches...
		# 
		# build the last version into the temp dir...
		my $somename = justFilename( $getlist[1] );
		$somename =~ s/\.patch$/.full/i;
		ilogStatus " - building " . justFilename( $somename ) . "\n";
		patch( $getlist[0], "$tempdir/$somename", $getlist[1] );
	
		# get the rest of the versions.
		for (my $x = 2; $x < @getlist; $x++)
		{
			if (!verifyChecksum( $getlist[$x], $getlist[$x] . ".md5sum" ))
			{
				ilogErrorAndStatus " - can't use patch file " . $getlist[$x] . ", checksum was invalid, possible corruption\n";
				return;
			}

			# if the date we're requesting is older than current file, 
			# get the last file
			if ($requested_date lt getDateValueForFile( "$tempdir/$somename" ))
			{
				$someoldname = $somename;
				$somename = justFilename( $getlist[$x] );
				$somename =~ s/\.patch$/.full/i;
				ilogStatus " - building " . justFilename( $somename ) . "\n";
				patch( "$tempdir/$someoldname", "$tempdir/$somename", $getlist[$x] );
				unlink( "$tempdir/$someoldname" );
			}
		}
		if (verifyChecksum( "$tempdir/$somename", "$dblocation/$somename.md5sum" ))
		{
			ilogStatus " - getting " . justFilename( $destlocation ) . "\n";
			copy( "$tempdir/$somename", $destlocation );
		}
		else
		{
			ilogErrorAndStatus " - [fail] can't get file, checksum was invalid, patches were fine, \n";
			ilogErrorAndStatus " -       something wronng during incremental patching process used \n";
			ilogErrorAndStatus " -       to restore file.\n";
         return;
		}
	}

	# otherwise, the file didn't exist at the specified date.
	else
	{
		ilogStatus " - [warning] date too old, file didn't exist.\n";
      ilogError " - [warning] date $requested_date too old, file $dblocation didn't exist.\n";
	}
}

if ($cmd_mode eq "backup")
{
	ilogErrorAndStatus "Backing up set[$cmd_set]....\n";	
	iprint_inc();
	if ($cmd_set eq "all")
	{
		foreach my $item (keys %sets)
		{
			ilogErrorAndStatus " - Backing up set[$item]....\n";	
			iprint_inc();
			backup( $item );
			iprint_dec();
		}
	}
	else
	{
		backup( $cmd_set );
	}
	iprint_dec();
}
elsif ($cmd_mode eq "list")
{
	if ($cmd_set eq "all")
	{
		iprint "Showing available sets, use -ll to list a set\n";
		foreach my $item (keys %sets)
		{
			iprint " \"$item\"\n";
			#ilogStatus " - Showing contents of set[$item]....\n";
			#iprint_inc();
			#list( $item );
			#iprint_dec();
		}
	}
	else
	{
		iprint "Showing contents of set[$cmd_set]....\n";	
		list( $cmd_set );
	}
}
elsif ($cmd_mode eq "verify")
{
	ilogStatus "Verifying integrity of set[$cmd_set]....\n";
	if ($cmd_set eq "all")
	{
		foreach my $item (keys %sets)
		{
			ilogStatus " - Verifying integrity of set[$item]....\n";
			iprint_inc();
			integrity( $item );
			iprint_dec();
		}
	}
	else
	{
		integrity( $cmd_set );
	}
}
elsif ($cmd_mode eq "get")
{
	ilogStatus "Copy from [$cmd_set]....\n";
	my $dblocation = $arg1;
	my $destlocation = $arg2;	

	$set = $cmd_set;
	my $dest = $sets{$set}->{"dest"};
	
	if (isDbFile( "$dest/$dblocation" ))
	{
		print " - getting '$dest/$dblocation'...\n";
		retrieve( "$dest/$dblocation", $destlocation, interpretDate( $arg3 ) );
	}
   else
	{
		print " - doing recursive copy of '$dest/$dblocation'...\n";
		retrieveDir( "$dest/$dblocation", $destlocation, interpretDate( $arg3 ) );
	}
}
elsif ($cmd_mode eq "help")
{
	usageInfo();
}
elsif ($cmd_mode eq "ver")
{
	usageInfo();
   ilogStatus "\nkbk Version == $version\n";
}
